@@ -1,10 +1,14 @@
 !-----------------------------------------------------------------------
 SUBROUTINE lr_apply_liouvillian( evc1, evc1_new, sevc1_new, interaction )
   !---------------------------------------------------------------------
-  ! ... applies linear response operator to response wavefunctions
-  ! OBM: or to be more exact this function is responsible for calculating L.q(i) and (L^T).p(i)
-  ! where q is evc1 and return partial qdash(i+1) and pdash(i+1) in evc1_new . Ultrasoft additions are
-  ! handled here...
+  ! Applies linear response operator to response wavefunctions
+  !   (H - E)*psi(k+q) + HXC   
+  ! 
+  ! Or to be more exact this routine is responsible for calculating
+  !    L.q(i) and (L^T).p(i)
+  ! Where q is evc1
+  ! It returns partial qdash(i+1) and pdash(i+1) in evc1_new.
+  !
   ! interaction=.true. corresponds to eq.(32)
   ! interaction=.false. corresponds to eq.(33)
   ! in Ralph Gebauer, Brent Walker  J. Chem. Phys., 127, 164106 (2007)
@@ -16,13 +20,14 @@ SUBROUTINE lr_apply_liouvillian( evc1, evc1_new, sevc1_new, interaction )
   USE cell_base,            ONLY : tpiba2
   USE fft_base,             ONLY : dffts
   USE fft_interfaces,       ONLY : fwfft
-  USE gvecs,              ONLY : nls, nlsm
+  USE gvecs,                ONLY : nls, nlsm
   USE gvect,                ONLY : nl, ngm, gstart, g, gg
   USE fft_base,             ONLY : dfftp
   USE io_global,            ONLY : stdout
   USE kinds,                ONLY : dp
   USE klist,                ONLY : nks, xk
-  USE lr_variables,         ONLY : evc0, revc0, rho_1, lr_verbosity, ltammd, size_evc, no_hxc
+  USE lr_variables,         ONLY : evc0, revc0, rho_1, lr_verbosity,&
+                                 & ltammd, size_evc, no_hxc 
   USE realus,               ONLY : igk_k,npw_k
   USE lsda_mod,             ONLY : nspin
   USE uspp,                 ONLY : vkb, nkb, okvan
@@ -30,91 +35,128 @@ SUBROUTINE lr_apply_liouvillian( evc1, evc1_new, sevc1_new, interaction )
   USE wavefunctions_module, ONLY : psic
   USE wvfct,                ONLY : nbnd, npwx, igk, g2kin, et
   USE control_flags,        ONLY : gamma_only
-  USE realus,               ONLY : real_space, fft_orbital_gamma, initialisation_level, &
-                                   bfft_orbital_gamma, calbec_rs_gamma, add_vuspsir_gamma, &
-                                   v_loc_psir, s_psir_gamma, real_space_debug, &
-                                   betasave, box_beta, maxbox_beta,newq_r
-  USE lr_variables,   ONLY : lr_verbosity
-  USE io_global,      ONLY : stdout
-  USE dfunct,         ONLY : newq
-  USE control_flags,  ONLY : tqr
-
-
+  USE realus,               ONLY : real_space, fft_orbital_gamma,&
+                                   & initialisation_level,&
+                                   & bfft_orbital_gamma,&
+                                   & calbec_rs_gamma,&
+                                   & add_vuspsir_gamma, v_loc_psir,&
+                                   & s_psir_gamma, real_space_debug,&
+                                   & betasave, box_beta, maxbox_beta,&
+                                   & newq_r 
+  USE lr_variables,         ONLY : lr_verbosity
+  USE io_global,            ONLY : stdout
+  USE dfunct,               ONLY : newq
+  USE control_flags,        ONLY : tqr
+  !
   !
   IMPLICIT NONE
   !
   COMPLEX(kind=dp),INTENT(in)  :: evc1(npwx,nbnd,nks)
-  COMPLEX(kind=dp),INTENT(out) :: evc1_new(npwx,nbnd,nks), sevc1_new(npwx,nbnd,nks)
+  COMPLEX(kind=dp),INTENT(out) :: evc1_new(npwx,nbnd,nks),&
+                               & sevc1_new(npwx,nbnd,nks)
+  ! output : sevc1_new = S * evc1_new
+  !
   LOGICAL, INTENT(in) :: interaction
   !
   !   Local variables
   !
   INTEGER :: ir, ibnd, ik, ig, ia, mbia
   INTEGER :: ijkb0, na, nt, ih, jh, ikb, jkb, iqs,jqs
-  real(kind=dp), ALLOCATABLE :: dvrs(:,:), dvrss(:)
-  COMPLEX(kind=dp), ALLOCATABLE :: dvrs_temp(:,:) !OBM This waste of memory was already there in lr_dv_of_drho
-  real(kind=dp), ALLOCATABLE :: d_deeq(:,:,:,:)
+  !
+  REAL(kind=dp), ALLOCATABLE :: dvrs(:,:), dvrss(:)
+  REAL(kind=dp), ALLOCATABLE :: d_deeq(:,:,:,:)
+  !
+  COMPLEX(kind=dp), ALLOCATABLE :: dvrs_temp(:,:)   
   COMPLEX(kind=dp), ALLOCATABLE :: spsi1(:,:)
+  !
   COMPLEX(kind=dp) :: fp, fm
-  REAL(DP), ALLOCATABLE, DIMENSION(:) :: w1, w2
+  !
+  REAL(kind=dp), ALLOCATABLE, DIMENSION(:) :: w1, w2
+  !
   !
   IF (lr_verbosity > 5) THEN
     WRITE(stdout,'("<lr_apply_liouvillian>")')
   ENDIF
   !
-
   CALL start_clock('lr_apply')
-  IF (interaction) CALL start_clock('lr_apply_int')
+  !
+  IF (interaction)      CALL start_clock('lr_apply_int')
   IF (.not.interaction) CALL start_clock('lr_apply_no')
   !
   ALLOCATE( d_deeq(nhm, nhm, nat, nspin) )
   d_deeq(:,:,:,:)=0.0d0
+  !
   ALLOCATE( spsi1(npwx, nbnd) )
   spsi1(:,:)=(0.0d0,0.0d0)
   !
-  IF( interaction ) THEN !If true, the full L is calculated
+  IF( interaction ) THEN 
+     !
+     ! Calculate the full L
+     !
      ALLOCATE( dvrs(dfftp%nnr, nspin) )
      ALLOCATE( dvrss(dffts%nnr) )
      dvrs(:,:)=0.0d0
      dvrss(:)=0.0d0
      !
+     ! Calculation of the charge density response
+     !
      CALL lr_calc_dens( evc1, .false. )
      !
+     ! Given the change of the charge density
+     ! we calculate the change of the Hartree and XC potential and
+     ! put it in dvrss()
+     !
      IF (no_hxc) THEN
-     !OBM no_hxc controls the hartree excange correlation addition, if true, they are not added
-      dvrs(:,1)=0.0d0
-      CALL interpolate (dvrs(:,1),dvrss,-1)
+        !
+        ! With no_hxc=.true. we recover the independent electrion 
+        ! approximation, so we zero the interation.
+        !
+        dvrs(:,1)=0.0d0
+        CALL interpolate (dvrs(:,1),dvrss,-1)
+        !
      ELSE
-       dvrs(:,1)=rho_1(:,1)
-       !
-       !call lr_dv_of_drho(dvrs)
-       ALLOCATE( dvrs_temp(dfftp%nnr, nspin) )
-       dvrs_temp = CMPLX( dvrs, 0.0d0, kind=dp )         
-       ! OBM: This memory copy was hidden in lr_dv_of_drho, can it be avoided?
-       DEALLOCATE ( dvrs )
-       CALL dv_of_drho(0,dvrs_temp,.false.)
-       ALLOCATE ( dvrs(dfftp%nnr, nspin) ) !SJB Worth getting rid of this memory bottle neck for the moment.
-       dvrs=dble(dvrs_temp)
-       DEALLOCATE(dvrs_temp)
-       !
-       IF ( okvan )  THEN
-        IF ( tqr ) THEN
-         CALL newq_r(dvrs,d_deeq,.true.)
-        ELSE
-         ALLOCATE( psic(dfftp%nnr) )
-         psic(:)=(0.0d0,0.0d0)
-         CALL newq(dvrs,d_deeq,.true.)
-         DEALLOCATE( psic )
+        !
+        dvrs(:,1)=rho_1(:,1)
+        !
+        ! In the gamma_only case dvrs is real, but dv_of_drho expects
+        ! a complex array on input, hence this temporary variable.
+        !
+        ALLOCATE( dvrs_temp(dfftp%nnr, nspin) )
+        dvrs_temp = CMPLX( dvrs, 0.0d0, kind=dp )         
+        !
+        DEALLOCATE ( dvrs )
+        !
+        CALL dv_of_drho(0,dvrs_temp,.FALSE.)
+        !
+        ALLOCATE ( dvrs(dfftp%nnr, nspin) ) 
+        dvrs=DBLE(dvrs_temp)
+        DEALLOCATE(dvrs_temp)
+        !
+        IF ( okvan )  THEN
+           IF ( tqr ) THEN
+              CALL newq_r(dvrs,d_deeq,.TRUE.)
+           ELSE
+              ALLOCATE( psic(dfftp%nnr) )
+              psic(:)=(0.0d0,0.0d0)
+              !
+              CALL newq(dvrs,d_deeq,.TRUE.)
+              !
+              DEALLOCATE( psic )
+           ENDIF
         ENDIF
-       ENDIF
-       CALL add_paw_to_deeq(d_deeq)
-       !
-       CALL interpolate (dvrs(:,1),dvrss,-1)
+        CALL add_paw_to_deeq(d_deeq)
+        !
+        ! Put the nteraction on the smooth grid.
+        !
+        CALL interpolate (dvrs(:,1),dvrss,-1)
      ENDIF
      !
   ENDIF
   !
+  ! Make sure the psic workspace is availible
+  !
   ALLOCATE ( psic (dfftp%nnr) )
+  !
   IF( gamma_only ) THEN
      !
      CALL lr_apply_liouvillian_gamma()
@@ -124,44 +166,48 @@ SUBROUTINE lr_apply_liouvillian( evc1, evc1_new, sevc1_new, interaction )
      CALL lr_apply_liouvillian_k()
      !
   ENDIF
+  !
   DEALLOCATE ( psic )
   !
   IF ( interaction .and. (.not.ltammd) ) THEN
      !
      !   Normal interaction
      !
-     WRITE(stdout,'(5X,"lr_apply_liouvillian: applying interaction: normal")')
+     WRITE(stdout,'(5X,"lr_apply_liouvillian: &
+          &applying interaction: normal")')
      !
-     !   Here evc1_new contains the interaction
+     ! Here we add the two terms: 
+     ! [H(k) - E(k)] * evc1(k)  +  P_c(k) [dV_{HXC} * revc0(k)]
      !
-     !OBM, blas
-     !sevc1_new=sevc1_new+(1.0d0,0.0d0)*evc1_new
-     CALL zaxpy(size_evc,cmplx(1.0d0,0.0d0,kind=dp),evc1_new(:,:,:),1,sevc1_new(:,:,:),1)
+     CALL zaxpy(size_evc,CMPLX(1.0d0,0.0d0,kind=dp),&
+          &evc1_new(:,:,:), 1, sevc1_new(:,:,:), 1)
      !
      !
   ELSEIF ( interaction .and. ltammd ) THEN
      !
      !   Tamm-dancoff interaction
      !
-     WRITE(stdout,'(5X,"lr_apply_liouvillian: applying interaction: tamm-dancoff")')
+     WRITE(stdout,'(5X,"lr_apply_liouvillian:&
+          & applying interaction: tamm-dancoff")')
      !
      !   Here evc1_new contains the interaction
      !
-     !OBM, blas
-     !sevc1_new=sevc1_new+(0.50d0,0.0d0)*evc1_new
-     CALL zaxpy(size_evc,cmplx(0.5d0,0.0d0,kind=dp),evc1_new(:,:,:),1,sevc1_new(:,:,:),1)
+     CALL zaxpy(size_evc,CMPLX(0.5d0,0.0d0,kind=dp),&
+          &evc1_new(:,:,:) , 1 , sevc1_new(:,:,:),1)
      !
      !
   ELSE
      !
      !   Non interacting
      !
-     WRITE(stdout,'(5X,"lr_apply_liouvillian: not applying interaction")')
+     WRITE(stdout,'(5X,"lr_apply_liouvillian:&
+          & not applying interaction")') 
      !
   ENDIF
   !
-  IF (gstart == 2 .and. gamma_only ) sevc1_new(1,:,:)=cmplx(real(sevc1_new(1,:,:),dp),0.0d0,dp)
-  ! (OBM: Why there is this check?)
+  IF (gstart == 2 .AND. gamma_only ) sevc1_new(1,:,:) = &
+       &CMPLX( REAL( sevc1_new(1,:,:), dp ), 0.0d0 ,dp )
+  !
   IF(gstart==2 .and. gamma_only) THEN
      !
      DO ik=1,nks
@@ -183,9 +229,19 @@ SUBROUTINE lr_apply_liouvillian( evc1, evc1_new, sevc1_new, interaction )
      !
   ENDIF
   !
+  ! Here we apply the S^{-1} operator.
+  ! See equations after Eq.(47) of B. Walker et al., J. Chem. Phys.
+  !  127, 164106 (2007).  
+  !
+  ! S^{-1} ( H(k)*evc1(k) - E(k) * S evc1(k) ) 
+  ! or
+  ! S^{-1} ( H(k)*evc1(k) - E(k) * S evc1(k)  +  P_c(k) [dV_{HXC} *
+  !  revc0(k)] ) 
+  !
   DO ik=1,nks
      !
-     CALL sm1_psi(.false.,ik,npwx,npw_k(ik),nbnd,sevc1_new(1,1,ik),evc1_new(1,1,ik))
+     CALL sm1_psi(.FALSE., ik, npwx, npw_k(ik), nbnd, &
+          &sevc1_new(1,1,ik), evc1_new(1,1,ik))
      !
   ENDDO
   !
@@ -194,10 +250,9 @@ SUBROUTINE lr_apply_liouvillian( evc1, evc1_new, sevc1_new, interaction )
   DEALLOCATE(d_deeq)
   DEALLOCATE(spsi1)
   !
-9000 FORMAT(/5x,'lr_apply_liouvillian: ibnd=',1X,i2,1X,'sevc1_new(G=0)[',i1,']',2(1x,e12.5)/)
   !
   CALL stop_clock('lr_apply')
-  IF (interaction) CALL stop_clock('lr_apply_int')
+  IF (interaction)      CALL stop_clock('lr_apply_int')
   IF (.not.interaction) CALL stop_clock('lr_apply_no')
   !
   RETURN
@@ -206,10 +261,19 @@ CONTAINS
   !
   SUBROUTINE lr_apply_liouvillian_gamma()
     !
-    !use becmod,              only : bec_type,becp
-    USE lr_variables,        ONLY : becp1
+    USE lr_variables,             ONLY : becp1, tg_revc0
+    USE wavefunctions_module,     ONLY : psic
+    USE realus,                   ONLY : tg_psic
+    USE mp_global,                ONLY : me_bgrp
+    USE fft_base,                 ONLY : dffts, tg_gather
     !
-    real(kind=dp), ALLOCATABLE :: becp2(:,:)
+    REAL(kind=dp), ALLOCATABLE :: becp2(:,:)
+    REAL(kind=dp), ALLOCATABLE :: tg_dvrss(:)
+    LOGICAL :: use_tg
+    INTEGER :: v_siz, incr, ioff
+    !
+    use_tg=dffts%have_task_groups
+    incr = 2
     !
     IF ( nkb > 0 .and. okvan ) THEN
        !
@@ -264,26 +328,49 @@ CONTAINS
           ENDDO
           !end: calculation of becp2
        ENDIF
+
+      IF( dffts%have_task_groups ) THEN
+         !
+         v_siz =  dffts%tg_nnr * dffts%nogrp
+         !
+         incr = 2 * dffts%nogrp
+         !
+         ALLOCATE( tg_dvrss(1:v_siz) )
+         tg_dvrss=0.0d0
+         !
+         CALL tg_gather(dffts, dvrss, tg_dvrss)
+         !
+      ENDIF
        !
        !   evc1_new is used as a container for the interaction
        !
        evc1_new(:,:,:)=(0.0d0,0.0d0)
        !
-       DO ibnd=1,nbnd,2
+       DO ibnd=1,nbnd,incr
           !
-          !   Product with the potential vrs = (vltot+vr)
-          ! revc0 is on smooth grid. psic is used upto smooth grid
-          DO ir=1,dffts%nnr
+          ! Product with the potential vrs = (vltot+vr)
+          ! revc0 is on smooth grid. psic is used up to smooth grid
+          !
+          IF(dffts%have_task_groups) THEN
              !
-             psic(ir)=revc0(ir,ibnd,1)*cmplx(dvrss(ir),0.0d0,dp)
+             DO ir=1, dffts%nr1x*dffts%nr2x*dffts%tg_npp( me_bgrp + 1 )
+                !
+                tg_psic(ir)=tg_revc0(ir,ibnd,1)*CMPLX(tg_dvrss(ir),0.0d0,dp)
+                !
+             ENDDO
              !
-          ENDDO
-
+          ELSE
+             !
+             DO ir=1,dffts%nnr
+                !
+                psic(ir)=revc0(ir,ibnd,1)*CMPLX(dvrss(ir),0.0d0,dp)
+                !
+             ENDDO
+             !
+          ENDIF
           !
-          !print *,"1"
           IF (real_space_debug > 7 .and. okvan .and. nkb > 0) THEN
           !THE REAL SPACE PART (modified from s_psi)
-                  !print *, "lr_apply_liouvillian:Experimental interaction part not using vkb"
                   !fac = sqrt(omega)
                   !
                   ijkb0 = 0
@@ -297,7 +384,6 @@ CONTAINS
                       IF ( ityp(ia) == nt ) THEN
                         !
                         mbia = maxbox_beta(ia)
-                        !print *, "mbia=",mbia
                         ALLOCATE( w1(nh(nt)),  w2(nh(nt)) )
                         w1 = 0.D0
                         w2 = 0.D0
@@ -308,7 +394,8 @@ CONTAINS
                           !
                           jkb = ijkb0 + jh
                           w1(ih) = w1(ih) + becp2(jkb, ibnd)
-                          IF ( ibnd+1 <= nbnd ) w2(ih) = w2(ih) + becp2(jkb, ibnd+1)
+                          IF ( ibnd+1 <= nbnd ) w2(ih) = w2(ih) + &
+                               &becp2(jkb, ibnd+1)
                           !
                           ENDDO
                         !
@@ -323,7 +410,10 @@ CONTAINS
                           DO ir = 1, mbia
                           !
                            iqs = jqs + ir
-                           psic( box_beta(ir,ia) ) = psic(  box_beta(ir,ia) ) + betasave(ia,ih,ir)*cmplx( w1(ih), w2(ih) )
+                           psic( box_beta(ir,ia) ) = &
+                                &psic(  box_beta(ir,ia) ) + &
+                                &betasave(ia,ih,ir)*&
+                                &CMPLX( w1(ih), w2(ih) )
                           !
                           ENDDO
                         !
@@ -337,25 +427,26 @@ CONTAINS
                     !
                     ENDDO
                   !
-                  ENDDO
+                 ENDDO
 
           ENDIF
-          !print *,"2"
           !
-          !   Back to reciprocal space This part is equivalent to bfft_orbital_gamma
+          !   Back to reciprocal space 
           !
           CALL bfft_orbital_gamma (evc1_new(:,:,1), ibnd, nbnd,.false.)
-          !print *,"3"
+          !
        ENDDO
        !
+       IF(dffts%have_task_groups) DEALLOCATE (tg_dvrss)
+       !
        !
        IF( nkb > 0 .and. okvan .and. real_space_debug <= 7) THEN
          !The non real_space part
           CALL dgemm( 'N', 'N', 2*npw_k(1), nbnd, nkb, 1.d0, vkb, &
                2*npwx, becp2, nkb, 1.d0, evc1_new, 2*npwx )
-          !print *, "lr_apply_liouvillian:interaction part using vkb"
           !
        ENDIF
+       !
        CALL stop_clock('interaction')
        !
     ENDIF
@@ -380,7 +471,8 @@ CONTAINS
     !
     DO ibnd=1,nbnd
        !
-       CALL zaxpy(npw_k(1), cmplx(-et(ibnd,1),0.0d0,dp), spsi1(:,ibnd), 1, sevc1_new(:,ibnd,1), 1)
+       CALL zaxpy(npw_k(1), CMPLX(-et(ibnd,1),0.0d0,dp), &
+            &spsi1(:,ibnd), 1, sevc1_new(:,ibnd,1), 1)
        !
     ENDDO
     !
@@ -392,12 +484,11 @@ CONTAINS
   !
   SUBROUTINE lr_apply_liouvillian_k()
     !
-    !use becmod,              only : becp
     USE lr_variables,        ONLY : becp1_c
     !
     COMPLEX(kind=dp), ALLOCATABLE :: becp2(:,:)
     !
-    IF( nkb > 0 .and. okvan ) THEN
+    IF( nkb > 0 .AND. okvan ) THEN
        !
        ALLOCATE(becp2(nkb,nbnd))
        becp2(:,:)=(0.0d0,0.0d0)
